# Copyright (C) 2011 Johan Rydberg
# Copyright (C) 2010 Bob Potter
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from txgossip.detector import FailureDetector


class PeerState(object):

    def __init__(self, clock, participant, name=None, PHI=8):
        self.clock = clock
        self.participant = participant
        self.max_version_seen = 0
        self.attrs = {}
        self.detector = FailureDetector()
        self.alive = False
        self.heart_beat_version = 0
        self.name = name
        self.PHI = PHI

    def set_name(self, name):
        self.name = name

    def update_with_delta(self, k, v, n):
        """."""
        # It's possibly to get the same updates more than once if
        # we're gossiping with multiple peers at once ignore them
        if n > self.max_version_seen:
            self.max_version_seen = n
            self.set_key(k,v,n)
            if k == '__heartbeat__':
                self.detector.add(self.clock.seconds())

    def update_local(self, k, v):
        # This is used when the peerState is owned by this peer
        self.max_version_seen += 1
        self.set_key(k, v, self.max_version_seen)

    def __iter__(self):
        return iter(self.attrs)

    def __len__(self):
        return len(self.attrs)

    def __contains__(self, key):
        return key in self.attrs

    def __setitem__(self, key, value):
        self.update_local(key, value)

    def set(self, key, value):
        self.update_local(key, value)

    def __getitem__(self, key):
        return self.attrs[key][0]

    def get(self, key, default=None):
        if key in self.attrs:
            return self.attrs[key][0]
        return default

    def has_key(self, key):
        return key in self.attrs

    def keys(self):
        return self.attrs.keys()

    def items(self):
        for k, (v, n) in self.attrs.items():
            yield k, v

    def set_key(self, k, v, n):
        self.attrs[k] = (v, n)
        self.participant.value_changed(self, str(k), v)

    def beat_that_heart(self):
        self.heart_beat_version += 1
        self.update_local('__heartbeat__', self.heart_beat_version);

    def deltas_after_version(self, lowest_version):
        """
        Return sorted by version.
        """
        deltas = []
        for key, (value, version) in self.attrs.items():
            if version > lowest_version:
                deltas.append((key, value, version))
        deltas.sort(key=lambda kvv: kvv[2])
        return deltas

    def check_suspected(self):
        phi = self.detector.phi(self.clock.seconds())
        if phi > self.PHI or phi == 0:
            self.mark_dead()
            return True
        else:
            self.mark_alive()
            return False

    def mark_alive(self):
        alive, self.alive = self.alive, True
        if not alive:
            self.participant.peer_alive(self)

    def mark_dead(self):
        if self.alive:
            self.alive = False
            self.participant.peer_dead(self)



# Copyright (C) 2011 Johan Rydberg
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import fnmatch


class LeaderElectionMixin:
    """Mixin for leader election among the nodes in the cluster.

    The C{value_changed} method should be called when any of the
    related keys are changed in the key-stores.

    The mixin will call C{leader_elected} when an election has taken
    place.

    When a node peforms an election, which it should do after any
    group change, it picks the node with the highest C{prio-key}
    value.  It updates its C{vote-key} with the name of that peer.
    When all peers have voted on the same peer, they one by one set
    their C{leader-key} to the same value as their C{vote-key}.  The
    election is over when all peers have reached a consensus on the
    C{leader-key} value.
    """

    PRIO_KEY = 'leader:priority'
    VOTE_KEY = 'leader:vote'
    LEADER_KEY = 'leader:leader'

    def __init__(self, clock, vote_delay=5):
        self.clock = clock
        self._election_timeout = None
        self._gossiper = None
        self.is_leader = None
        self.vote_delay = vote_delay

    def make_connection(self, gossiper):
        self._gossiper = gossiper
        self.start_election()

    def _check_consensus(self, key):
        """Check if all peers have the same value for C{key}.

        Return the value if they all have the same value, otherwise
        return C{None}.
        """
        correct = self._gossiper.get(key)
        for peer in self._gossiper.live_peers:
            if not key in peer.keys():
                return None
            value = peer.get(key)
            if value != correct:
                return None
        return correct

    def value_changed(self, peer, key, value):
        """Inform about a change of a key-value pair.

        @param peer: The peer that changed a value.
        @param key: The key.
        @param value: The new value.

        @return: C{True} if this method acted on the change, or if it
           was unrelated.
        """
        if key == self.VOTE_KEY:
            leader = self._check_consensus(self.VOTE_KEY)
            if leader:
                self._gossiper.set(self.LEADER_KEY, leader)
        elif key == self.LEADER_KEY:
            leader = self._check_consensus(self.LEADER_KEY)
            if leader:
                self.leader_elected(
                    self._gossiper.name == leader, leader)
        elif key == self.PRIO_KEY:
            self.start_election()

        return key in (self.VOTE_KEY, self.LEADER_KEY, self.PRIO_KEY)

    def _vote(self):
        """Perform an election."""
        self._election_timeout = None

        # Check if we're one of the peers that would like to become
        # master.
        vote = None
        curr = self._gossiper.get(self.PRIO_KEY)
        if curr is not None:
            vote = self._gossiper

        for peer in self._gossiper.live_peers:
            prio = peer.get(self.PRIO_KEY)
            if prio is None:
                # This peer did not want to become leader.
                continue
            elif curr is None:
                curr = prio
                vote = peer
            elif prio > curr:
                curr = prio
                vote = peer
            elif prio == curr:
                # We need to break the tie.
                if hash(peer.name) > hash(vote.name):
                    vote = peer
        self._gossiper.set(self.VOTE_KEY, vote.name)

    def start_election(self):
        """Start an election.

        Elections should be started when the cluster membership view
        changes (i.e.g, when a peer joins the cluster, or when a peer
        dies).

        It is safe to call this while an election is taking place.
        """
        if self._election_timeout is not None:
            self._election_timeout.cancel()
        self._election_timeout = self.clock.callLater(self.vote_delay,
                self._vote)

    def leader_elected(self, is_leader, leader):
        """Notifcation about leader election result.

        @param is_leader: C{True} if this peer is the leader.
        @param leader: The address of the peer that is the leader.
        """
        self.is_leader = is_leader

    def peer_dead(self, peer):
        """A peer is dead."""
        self.start_election()

    def peer_alive(self, peer):
        """A peer is alive."""
        self.start_election()


class KeyStoreMixin:
    """Mixin that implements a distributed key-value store."""

    def __init__(self, clock, storage, ignore_keys=[]):
        """Initialize key-value store mixin.

        @param clock: Something that can report the time, normally a
            Twisted reactor.
        @param storage: A backing storage object that responds to the
            normal C{dict}-like protocol.
        @param ignore_keys: A sequence of keys that should not be
            replicated between the peers.
        """
        self.clock = clock
        self._storage = storage
        self._ignore_keys = ignore_keys
        self._gossiper = None

    def make_connection(self, gossiper):
        self._gossiper = gossiper

    def persist_key_value(self, key, timestamped_value):
        self._storage[key] = timestamped_value
        if hasattr(self._storage, 'sync'):
            self._storage.sync()

    def replicate_key_value(self, peer, key, timestamped_value):
        timestamp, value = timestamped_value
        if key in self._storage:
            current_timestamp, current_value = self._storage[key]
            if timestamp <= current_timestamp:
                return
        # We replicate the value.
        self._gossiper.set(key, timestamped_value)

    def value_changed(self, peer, key, timestamp_value):
        """A peer has changed its value."""
        if key == '__heartbeat__' or key in self._ignore_keys:
            return
        if peer.name == self._gossiper.name:
            self.persist_key_value(key, timestamp_value)
        else:
            self.replicate_key_value(peer, key, timestamp_value)

    def set(self, key, value):
        self._gossiper.set(key, [self.clock.seconds(), value])

    def __setitem__(self, key, value):
        self.set(key, value)

    def __getitem__(self, key):
        return self._gossiper.get(key)[1]

    def get(self, key, default=None):
        if key in self.keys():
            return self[key]
        return default

    def keys(self, pattern=None):
        """Return a iterable of all available keys."""
        if pattern is None:
            return self._gossiper.keys()
        else:
            keys = self._gossiper.keys()
            return [key for key in keys
                    if fnmatch.fnmatch(key, pattern)]

    def load_from(self, storage):
        for key in storage:
            if key not in self._ignore_keys:
                self._gossiper.set(key, storage[key])

    def __contains__(self, key):
        return key in self.keys()

    def peer_dead(self, peer):
        """A peer is dead."""

    def peer_alive(self, peer):
        """A peer is alive."""



# Copyright (C) 2011 Johan Rydberg
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mockito import mock, when, verify

from twisted.trial import unittest
from twisted.internet import task

from txgossip.recipies import KeyStoreMixin


class KeyStoreTestCase(unittest.TestCase):
    """Test cases for the key-value store mixin."""

    def setUp(self):
        self.clock = task.Clock()
        self.storage = {}
        self.keystore = KeyStoreMixin(self.clock, self.storage)
        self.gossiper = mock()
        self.gossiper.name = 'self'
        self.keystore.make_connection(self.gossiper)

    def test_replicate_when_remote_peer_changed_value(self):
        self.keystore.value_changed('a', 'k', (0, 'value'))
        verify(self.gossiper).set('k', (0, 'value'))
        self.assertNotIn('k', self.storage)

    def test_ignore_replication_when_remote_peer_has_old_value(self):
        self.storage['k'] = (1, 'value')
        self.keystore.value_changed('a', 'k', (0, 'value'))
        verify(self.gossiper, times=0).set('k', (0, 'value'))

    def test_persist_value_when_set_on_local_peer(self):
        self.keystore.value_changed('self', 'k', (0, 'value'))
        self.assertIn('k', self.storage)
        self.assertEquals(self.storage['k'], (0, 'value'))

    def test_keys_returns_all_keys_in_gossiper(self):
        when(self.gossiper).keys().thenReturn(['a', 'b'])
        self.assertEquals(self.keystore.keys(), ['a', 'b'])

    def test_keys_can_be_filtered(self):
        when(self.gossiper).keys().thenReturn(['ab', 'ba'])
        self.assertEquals(self.keystore.keys('b*'), ['ba'])

    def test_contains_use_keys_from_gossiper(self):
        when(self.gossiper).keys().thenReturn(['a', 'b'])
        self.assertIn('a', self.keystore)

    def test_get_results_default_value_if_value_not_present(self):
        when(self.gossiper).keys().thenReturn([])
        self.assertEquals(self.keystore.get('a', '!'), '!')

    def test_get_returns_value_from_gossip_state(self):
        when(self.gossiper).keys().thenReturn(['a'])
        when(self.gossiper).get('a').thenReturn((0, '!'))
        self.assertEquals(self.keystore.get('a'), '!')



# Copyright (C) 2011 Johan Rydberg
# Copyright (C) 2010 Bob Potter
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math


class FailureDetector(object):

    def __init__(self):
        self.last_time = None
        self.intervals = []

    def add(self, arrival_time):
        last_time, self.last_time = self.last_time, arrival_time
        if last_time is None:
            i = 0.75
        else:
            i = arrival_time - last_time
        self.intervals.append(i)
        if len(self.intervals) > 1000:
            self.intervals.pop(0)

    def phi(self, current_time):
        if self.last_time is None:
            return 0
        current_interval = current_time - self.last_time
        exp = -1 * current_interval / self.interval_mean()
        return -1 * (math.log(pow(math.e, exp)) / math.log(10))

    def interval_mean(self):
        return sum(self.intervals) / float(len(self.intervals))



# Copyright (C) 2011 Johan Rydberg
# Copyright (C) 2010 Bob Potter
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random
import json

from txgossip.state import PeerState
from txgossip.scuttle import Scuttle
from twisted.python import log
from twisted.internet.protocol import DatagramProtocol
from twisted.internet import task


def _address_from_peer_name(name):
    address, port = name.split(':', 1)
    return address, int(port)

def _address_to_peer_name(address):
    return '%s:%d' % (address.host, address.port)


class Participant(object):
    """Base class for participants."""

    def make_connection(self, gossiper):
        """Attach this participant to a gossiper."""
        self.gossiper = gossiper

    def peer_alive(self, peer_name):
        """Report that there's a peer alive."""

    def peer_dead(self, peer_name):
        """Report that there's a peer dead."""


class Gossiper(DatagramProtocol):

    def __init__(self, clock, participant, address=None):
        """Create a new gossiper.

        @param address: Listen address if the gossiper will not be
            bound to a specific listen interface.
        @param address: C{str}
        """
        self.state = PeerState(clock, participant)
        self._states = {}
        self._address = address
        self._scuttle = Scuttle(self._states, self.state)
        self._heart_beat_timer = task.LoopingCall(self._beat_heart)
        self._heart_beat_timer.clock = clock
        self._gossip_timer = task.LoopingCall(self._gossip)
        self._gossip_timer.clock = clock
        self.clock = clock
        self.participant = participant
        self._seeds = []

    def _setup_state_for_peer(self, peer_name):
        """Setup state for a new peer."""
        self._states[peer_name] = PeerState(self.clock, self.participant,
            name=peer_name)

    def seed(self, seeds):
        """Tell this gossiper that there are gossipers to
        be found at the given endpoints.

        @param seeds: a sequence of C{'ADDRESS:PORT'} strings.
        """
        self._seeds.extend(seeds)
        self._handle_new_peers(seeds)

    def _handle_new_peers(self, names):
        """Set up state for new peers."""
        for peer_name in names:
            if peer_name in self._states:
                continue
            self._setup_state_for_peer(peer_name)

    def _determine_endpoint(self):
        """Determine the IP address of this peer.

        @raises Exception: If it is not impossible to figure out the
            address.
        @return: a C{ADDRESS:PORT} string.
        """
        # Figure our our endpoint:
        host = self.transport.getHost()
        if not self._address:
            self._address = host.host
            if self._address == '0.0.0.0':
                raise Exception("address not specified")
        return '%s:%d' % (self._address, host.port)

    def startProtocol(self):
        """Start protocol."""
        self.name = self._determine_endpoint()
        self.state.set_name(self.name)
        self._states[self.name] = self.state
        self._heart_beat_timer.start(1, now=True)
        self._gossip_timer.start(1, now=True)
        self.participant.make_connection(self)

    def stopProtocol(self):
        """Stop protocol."""
        self._gossip_timer.stop()
        self._heart_beat_timer.stop()

    def _beat_heart(self):
        """Beat heart of our own state."""
        self.state.beat_that_heart()

    def datagramReceived(self, data, address):
        """Handle a received datagram."""
        self._handle_message(json.loads(data), address)

    def _gossip(self):
        """Initiate a round of gossiping."""
        live_peers = self.live_peers
        dead_peers = self.dead_peers
        if live_peers:
            self._gossip_with_peer(random.choice(live_peers))

        prob = len(dead_peers) / float(len(live_peers) + 1)
        if random.random() < prob:
            self._gossip_with_peer(random.choice(dead_peers))

        for state in self._states.values():
            if state.name != self.name:
                state.check_suspected()

    def _gossip_with_peer(self, peer):
        """Send a gossip message to C{peer}."""
        self.transport.write(json.dumps({
            'type': 'request', 'digest': self._scuttle.digest()
            }), _address_from_peer_name(peer.name))

    def _handle_message(self, message, address):
        """Handle an incoming message."""
        if message['type'] == 'request':
            self._handle_request(message, address)
        elif message['type'] == 'first-response':
            self._handle_first_response(message, address)
        elif message['type'] == 'second-response':
            self._handle_second_response(message, address)

    def _handle_request(self, message, address):
        """Handle an incoming gossip request."""
        deltas, requests, new_peers = self._scuttle.scuttle(
            message['digest'])
        self._handle_new_peers(new_peers)
        response = json.dumps({
            'type': 'first-response', 'digest': requests, 'updates': deltas
            })
        self.transport.write(response, address)

    def _handle_first_response(self, message, address):
        """Handle the response to a request."""
        self._scuttle.update_known_state(message['updates'])
        response = json.dumps({
            'type': 'second-response',
            'updates': self._scuttle.fetch_deltas(
                    message['digest'])
            })
        self.transport.write(response, address)

    def _handle_second_response(self, message, address):
        """Handle the ack of the response."""
        self._scuttle.update_known_state(message['updates'])

    def live_peers():
        """Property for all peers that we know is alive.

        The property holds a sequence L{PeerState}'s.
        """
        def get(self):
            return [p for (n, p) in self._states.items()
               if p.alive and n != self.name]
        return get,
    live_peers = property(*live_peers())

    def dead_peers():
        """Property for all peers that we know is dead.

        The property holds a sequence L{PeerState}'s.
        """
        def get(self):
            return [p for (n, p) in self._states.items()
               if not p.alive and n != self.name]
        return get,
    dead_peers = property(*dead_peers())

    # dict-like interface:

    def __getitem__(self, key):
        return self.state[key]

    def set(self, key, value):
        self.state[key] = value

    def __setitem__(self, key, value):
        self.set(key, value)

    def __contains__(self, key):
        return key in self.state

    def has_key(self, key):
        return key in self.state

    def __len__(self):
        return len(self.state)

    def __iter__(self):
        return iter(self.state)

    def keys(self):
        return self.state.keys()

    def get(self, key, default=None):
        if key in self.state:
            return self.state[key]
        return default



# Copyright (C) 2011 Johan Rydberg
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mockito import mock, when, verify, verifyNoMoreInteractions

from twisted.trial import unittest
from twisted.internet import task

from txgossip.recipies import LeaderElectionMixin


class KeyStoreTestCase(unittest.TestCase):
    """Test cases for the key-value store mixin."""

    def setUp(self):
        self.clock = task.Clock()
        self.storage = {}
        self.election = LeaderElectionMixin(self.clock)
        self.gossiper = mock()
        self.gossiper.name = 'self'
        self.election.make_connection(self.gossiper)

    def test_make_connection_starts_election(self):
        self.assertTrue(self.election._election_timeout)

    def test_waits_for_consensus_on_vote_before_leader_elected(self):
        peer = mock()
        when(peer).get(self.election.VOTE_KEY).thenReturn('a')
        self.gossiper.live_peers = [peer]
        when(peer).keys().thenReturn([self.election.VOTE_KEY])
        when(self.gossiper).get(self.election.VOTE_KEY).thenReturn('b')
        self.election.value_changed('peer-a', self.election.VOTE_KEY,
                                    'a')
        verify(self.gossiper).get(self.election.VOTE_KEY)
        verify(peer).keys()
        verify(peer).get(self.election.VOTE_KEY)
        verifyNoMoreInteractions(self.gossiper)
        verifyNoMoreInteractions(peer)

    def test_leader_elected_when_consensus_reached_on_votes(self):
        peer = mock()
        when(peer).get(self.election.VOTE_KEY).thenReturn('a')
        when(peer).keys().thenReturn([self.election.VOTE_KEY])
        self.gossiper.live_peers = [peer]
        when(self.gossiper).get(self.election.VOTE_KEY).thenReturn('a')
        self.election.value_changed('peer-a', self.election.VOTE_KEY, 'a')
        verify(self.gossiper).get(self.election.VOTE_KEY)
        verify(self.gossiper).set(self.election.LEADER_KEY, 'a')
        verify(peer).keys()
        verify(peer).get(self.election.VOTE_KEY)
        verifyNoMoreInteractions(self.gossiper)
        verifyNoMoreInteractions(peer)

    def test_waits_for_consensus_on_leader_before_proclaiming(self):
        peer = mock()
        when(peer).get(self.election.LEADER_KEY).thenReturn('a')
        when(peer).keys().thenReturn([self.election.LEADER_KEY])
        self.gossiper.live_peers = [peer]
        when(self.gossiper).get(self.election.LEADER_KEY).thenReturn('b')
        self.election.value_changed('peer', self.election.LEADER_KEY, 'a')
        verify(self.gossiper).get(self.election.LEADER_KEY)
        verify(peer).keys()
        verify(peer).get(self.election.LEADER_KEY)
        verifyNoMoreInteractions(self.gossiper)
        verifyNoMoreInteractions(peer)

    def test_leader_proclaimed_when_consensus_reached_on_leader(self):
        peer = mock()
        when(peer).keys().thenReturn([self.election.LEADER_KEY])
        when(peer).get(self.election.LEADER_KEY).thenReturn('a')
        self.gossiper.live_peers = [peer]
        when(self.gossiper).get(self.election.LEADER_KEY).thenReturn('a')
        self.election.value_changed('peer', self.election.LEADER_KEY, 'a')
        self.assertEquals(self.election.is_leader, False)

    def test_sets_is_leader_when_leader_is_gossiper(self):
        self.gossiper.live_peers = []
        when(self.gossiper).get(self.election.LEADER_KEY).thenReturn('self')
        self.election.value_changed('self', self.election.LEADER_KEY, 'self')
        self.assertEquals(self.election.is_leader, True)

    def test_vote_for_peer_with_highest_priority(self):
        peer = mock()
        peer.name = 'peer'
        when(peer).get(self.election.PRIO_KEY).thenReturn(1)
        when(self.gossiper).get(self.election.PRIO_KEY).thenReturn(0)
        self.gossiper.live_peers = [peer]
        self.election._vote()
        verify(self.gossiper).set(self.election.VOTE_KEY, 'peer')

    def test_breaks_tie_using_hash_function(self):
        peer = mock()
        peer.name = 'peer'
        when(peer).get(self.election.PRIO_KEY).thenReturn(0)
        when(self.gossiper).get(self.election.PRIO_KEY).thenReturn(0)
        self.gossiper.live_peers = [peer]
        self.election._vote()
        verify(self.gossiper).set(self.election.VOTE_KEY, 'peer')
        self.assertTrue(hash(peer.name) > hash(self.gossiper.name))

    def test_ingores_peer_that_has_no_priority(self):
        peer = mock()
        peer.name = 'peer'
        when(peer).get(self.election.PRIO_KEY).thenReturn(None)
        when(self.gossiper).get(self.election.PRIO_KEY).thenReturn(0)
        self.gossiper.live_peers = [peer]
        self.election._vote()
        verify(self.gossiper).set(self.election.VOTE_KEY, 'self')

    def test_ingores_self_if_no_priority_set(self):
        peer = mock()
        peer.name = 'peer'
        when(peer).get(self.election.PRIO_KEY).thenReturn(0)
        when(self.gossiper).get(self.election.PRIO_KEY).thenReturn(None)
        self.gossiper.live_peers = [peer]
        self.election._vote()
        verify(self.gossiper).set(self.election.VOTE_KEY, 'peer')





# Copyright (C) 2011 Johan Rydberg
# Copyright (C) 2010 Bob Potter
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from txgossip.state import PeerState


class Scuttle(object):

    def __init__(self, peers, local_peer):
        self.peers = peers
        self.local_peer = local_peer

    def digest(self):
        digest = {}
        for peer, state in self.peers.items():
            digest[peer] = state.max_version_seen
        return digest

    def scuttle(self, digest):
        deltas_with_peer = []
        requests = {}
        new_peers = []
        for peer, digest_version in digest.items():
            if not peer in self.peers:
                requests[peer] = 0
                new_peers.append(peer)
            else:
                state = self.peers[peer]
                if state.max_version_seen > digest_version:
                    deltas_with_peer.append((
                            peer,
                            self.peers[peer].deltas_after_version(digest_version)
                            ))
                elif state.max_version_seen < digest_version:
                    requests[peer] = state.max_version_seen

        # Sort by peers with most deltas
        def sort_metric(a, b):
            return len(b[1]) - len(a[1])
        deltas_with_peer.sort(cmp=sort_metric)

        deltas = []
        for (peer, peer_deltas) in deltas_with_peer:
            for (key, value, version) in peer_deltas:
                deltas.append((peer, key, value, version))

        return deltas, requests, new_peers

    def update_known_state(self, deltas):
        for peer, key, value, version in deltas:
            self.peers[peer].update_with_delta(
                str(key), value, version)

    def fetch_deltas(self, requests):
        deltas = []
        for peer, version in requests.items():
            peer_deltas = self.peers[peer].deltas_after_version(
                version)
            for (key, value, version) in peer_deltas:
                deltas.append((peer, key, value, version))
        return deltas



from setuptools import setup, find_packages
setup(name='txgossip',
      version='0.1',
      description='simple cluster communication protocol',
      author='Johan Rydberg',
      author_email='johan.rydberg@gmail.com',
      url='http://github.com/jrydberg/txgossip',
      packages=find_packages()
)
